# Copyright 2023 Eurobios
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import os.path

import pandas as pd

home_directory = os.path.expanduser('~')

cache_folder = f'{home_directory}/.acrocord/cache'
cache_volume_gio = 0.5


def name(table_name: str, columns: iter = (), where: str = '',
         limit: int = 0) -> str:
    return f"{table_name}_{columns}_{where}_{limit}".replace(
        "(", "_").replace(")", "_")


def folder_size_gio():
    return sum(
        os.path.getsize(f"{cache_folder}/{f}")
        for f in os.listdir(cache_folder)) / 1e9


def set_cache_folder():
    if not os.path.exists(cache_folder):
        os.makedirs(cache_folder)


def write(data: pd.DataFrame, table_name, **kwargs):
    if folder_size_gio() < cache_volume_gio:
        data.to_pickle(f"{cache_folder}/{name(table_name, **kwargs)}")


def read(table_name: str, **kwarg):
    name_ = name(table_name, **kwarg)
    if name_ in os.listdir(cache_folder):
        data = pd.read_pickle(f'{cache_folder}/{name_}')
        return data
    else:
        return pd.DataFrame()


def clean_cache():
    if os.path.exists(cache_folder):
        for f in os.listdir(cache_folder):
            os.remove(f"{cache_folder}/{f}")



# Copyright 2023 Eurobios
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.#
from functools import wraps
from typing import Iterable, Dict

from sqlalchemy import text

from acrocord.misc import execution


def execute(cmd, hide=False, fetch=False):
    @wraps(cmd)
    def wrapper(*args, **kwargs):
        cmd_sql = cmd(*args, **kwargs)
        execute_sql_cmd(args[0], cmd_sql, hide=hide, fetch=fetch)

    return wrapper


def execute_sql_cmd(connection, cmd: str, hide=False, fetch=False):
    if not hide:
        execution.print_(cmd)
    with connection.engine.connect() as cursor:
        res = cursor.execute(text(cmd))
        cursor.commit()
        if fetch:
            ret = res.fetchall()
            return ret


def merge(connection, table1, table2, out="tmp_merge", out_type="VIEW",
          suffixe=("_x", "_y"), on="", left_on="", right_on=""):
    out = f"{connection._get_schema(out)}.{connection._get_name(out)}"
    out_save = out
    if out in [table1, table2]:
        out = f"{connection._get_schema(out)}.tmp_merge"
    if left_on == "":
        left_on = on
        right_on = on
    if isinstance(left_on, str):
        right_on, left_on = [right_on], [left_on]
    left_on, right_on = [l.lower() for l in left_on], [r.lower() for r in
                                                       right_on]

    all_columns = connection.get_columns(table1) + connection.get_columns(
        table2)
    col_intersect = set(connection.get_columns(table1)
                        ).intersection(connection.get_columns(table2))
    if suffixe != ("_x", "_y"):
        col_intersect = all_columns

    db_str = f"CREATE {out_type} {out} AS (\n  SELECT "
    for col in connection.get_columns(table1):
        db_str += table1 + "." + col
        if col in col_intersect and col not in left_on:
            db_str += " AS " + col + suffixe[0]
        db_str += ","
    for col in connection.get_columns(table2):
        if col not in left_on:
            db_str += table2 + "." + col
            if col in col_intersect:
                db_str += " AS " + col + suffixe[1]
            db_str += ","
    db_str = db_str[:-1]
    db_str += f" FROM {table1},{table2}"

    left_on_ = [f"{table1}.{l}" for l in left_on]
    right_on_ = [f"{table1}.{r}" for r in right_on]

    left_on_ = ", ".join(left_on_)
    right_on_ = ", ".join(right_on_)
    db_str += "\n  WHERE (" + left_on_ + ")=(" + right_on_ + "));"
    if connection.active_sql_printing:
        print(db_str)
    connection.drop_table(out, type_=out_type)
    execute_sql_cmd(connection, db_str, hide=True)

    if out_save in [table1, table2]:
        connection.drop_table(out_save)
        connection.replace(out, out_save, type_=out_type)


def count(table_name):
    sql_cmd = f"SELECT COUNT(*) FROM {table_name};"
    return sql_cmd


def drop_table(table_name, option="CASCADE", type_="TABLE"):
    return f"DROP {type_} IF EXISTS {table_name} {option};"


def list_table(schema: str):
    """

    Parameters
    ----------
    schema: str
        Schema name

    Returns
    -------
        SQL Query
    """
    sql_cmd = f"SELECT * FROM information_schema.tables " \
              f"WHERE table_schema = '{schema}'"
    return sql_cmd


def add(name, type_, column_name, dtype) -> str:
    return f"ALTER {type_} {name} ADD {column_name} {dtype}"


def drop(name: str, type_: str, column_name: str):
    """
    Drop a column or object in table

    Parameters
    ----------
    name: str
        Object to alter
    type_: str

        Type of object to alter
        - VIEW
        - TABLE (default)

    column_name: str
        Column or object to drop

    Returns
    -------
        SQL Query
    """
    return f"ALTER {type_} {name} DROP {column_name} "


def create_schema(name: str) -> str:
    """

    Parameters
    ----------
    name: str
        Name of the schema to create

    Returns
    -------
        SQL Query
    """
    return f"CREATE SCHEMA IF NOT EXISTS {name}"


def create_table(name: str, columns: Iterable[str], dtypes: Iterable[str],
                 column_comments: Dict[str, str] = None) -> str:
    """

    Parameters
    ----------
    name: str
        The name of the table
    columns: iterable of str
        The names of the columns
    dtypes: iterable of str
        The types of columns
    column_comments : dict
        The comment on columns `{column1: comment 1, ...}`

    Returns
    -------
        SQL Query
    """
    col_zip = [f'{" " * 4}"{c}"{" " * max(1, (15 - len(c)))}' + db_data_type(
        dtypes[i])
        for i, c in enumerate(columns)]
    sql_cmd = f"CREATE TABLE {name}( \n"
    sql_cmd += ",\n".join(col_zip) + "\n)"
    sql_cmd += '\n;'
    if column_comments is not None:
        for col_name, col_comment in column_comments.items():
            col_comment = col_comment.replace("'", "''")
            sql_cmd += f"COMMENT ON COLUMN {name}" \
                       f".{col_name} IS '{col_comment}';\n"
    return sql_cmd


_conversion_table = {"uint8": "int2",
                     "uint16": "int4",
                     "uint32": "int8",
                     "int8": "int2",
                     "Int8": "int2",
                     "int16": "int2",
                     "Int16": "int2",
                     "float16": "float4",
                     "float32": "float8",
                     "int32": "int4",
                     "Int32": "int4",
                     "int64": "int8",
                     "Int64": "int8",
                     "float64": "float8",
                     "Float64": "float8",
                     "float": "float8",
                     "bytes_": "text",
                     "object": "text",
                     "nan": "text",
                     "str": "text",
                     "string": "text",
                     "bool": "bool",
                     "boolean": "bool",
                     "datetime64": "timestamp",
                     "category": 'int4'}

_conversion_table_inv = {
    'int2': 'int16', 'float4': 'float16', 'int4': 'int32', 'int8': 'int64',
    'float8': 'float64', 'text': 'str', 'bool': 'bool',
    'datetime64': 'timestamp'}


def db_data_type(data_type: str, invert: bool = False) -> str:
    """

    Parameters
    ----------
    data_type: str
        data type to translate into postgresql one

    invert: bool
        reciprocal conversion

    Returns
    -------

    """
    if "datetime" in data_type:
        return "timestamp"
    if "timestamp" in data_type:
        return "datetime64[s]"
    data_type = str(data_type)

    if invert:
        return _conversion_table_inv[data_type]
    return _conversion_table[data_type]


def get_metadata(table_name: str) -> str:
    """

    Parameters
    ----------
    table_name: str
        Name of the table from which the metadata should be retrieved

    Returns
    -------
        SQL Query
    """
    schema, table = table_name.split('.')
    sql_cmd = f"""
        select
            table_cols.table_schema,
            table_cols.table_name,
            table_cols.column_name,
            pgd.description
        from pg_catalog.pg_statio_all_tables as st
        full outer join pg_catalog.pg_description pgd on (
            pgd.objoid = st.relid
        )
        full outer join information_schema.columns table_cols on (
            pgd.objsubid   = table_cols .ordinal_position and
            table_cols.table_schema = st.schemaname and
            table_cols.table_name   = st.relname
        )
        where
        table_cols.table_schema = '{schema}' and
        table_cols.table_name = '{table}' """
    return sql_cmd



# Copyright 2023 Eurobios
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.#

from datetime import datetime
from typing import Union
from typing import Optional
from acrocord.misc.execution import execution_time
from acrocord.misc.execution import print_
from acrocord.databasepg import ConnectDatabase

import warnings
import pandas as pd

warnings.filterwarnings('ignore')

__DEFAULT_COL_NAME__ = ['date', 'value', 'message', 'other_info']
__DEFAULT_COL_TYPE__ = ['datetime', 'str', 'str', 'str']


class Logger:

    def __init__(self, connection: ConnectDatabase, table_name: str, schema: str = 'public', columns=None,
                 column_types=None):
        """
        Initialize class logger to write log messages
        Create a new table to write logs if it doesn't already exist

        Parameters
        ----------
        connection: ConnectDatabase's instance
            The connection instance, allowing to connect to a specific database
        table_name: str
            The table name in the database
        schema: str
            The schema name in the database, default is 'public'
        columns: Iterable of str
            columns names
        column_types: Iterable if str
            The columns types
        """

        if column_types is None:
            column_types = __DEFAULT_COL_TYPE__
        if columns is None:
            columns = __DEFAULT_COL_NAME__

        self.db_connection = connection
        self.table_name = f"{schema}.{table_name}"
        print_(f"{self.table_name = }", color='OKBLUE')

        if table_name in self.db_connection.get_table_names(schema):
            print_(f"\nTable {table_name} ALREADY EXIST, do not create a new one", color='OKBLUE')
            columns = self.db_connection.get_columns(self.table_name)
            column_types = self.db_connection.get_dtypes(self.table_name)
        else:
            print_(f"CREATE new table with {table_name = }", color='OKGREEN')
            self.db_connection.create_table(name=self.table_name, columns=columns, dtypes=column_types)

        self.columns = columns
        self.column_types = column_types

    @execution_time
    def write_log(self, data: Union[pd.Series, pd.DataFrame, dict]) -> Optional[ValueError]:
        """
        Main function to write log messages

        Parameters
        ----------
        data: pd.Series or pd.Dataframe or dictionary
            The log to be written to the table, specifically:
                - pd.Series: all values are written into 'value' column
                - pd.DataFrame: only data belonging to predefined columns are written
                - dict: only one log (one value for each key) is written
        Returns
        -------
            None
        """

        if isinstance(data, pd.Series):
            return self._write_log(self._series2df(data))
        elif isinstance(data, dict):
            return self._write_log(self._dict2df(data))
        elif isinstance(data, pd.DataFrame):
            return self._write_log(data.copy())
        else:
            print_(f"\nError of data type = {type(data)}", color='WARNING')
            raise ValueError("Input data should be type of [pd.Series, pd.DataFrame, or dict]")

    @staticmethod
    def _series2df(data: pd.Series) -> pd.DataFrame:
        """
        Convert pd.Series into pd.Dataframe

        Parameters
        ----------
        data: pd.Series
            The value to be converted

        Returns
        -------
            pd.DataFrame
        """
        return data.to_frame(name='value')

    @staticmethod
    def _dict2df(data: dict) -> pd.DataFrame:
        """
        Convert a dictionary into pd.Dataframe

        Parameters
        ----------
        data: dict
            The data to be converted, only one value for each key

        Returns
        -------
            pd.DataFrame
        """
        return pd.DataFrame(data, index=[0])

    def _write_log(self, log_message: pd.DataFrame) -> None:
        """

        Parameters
        ----------
        log_message: pd.DataFrame
            The log message to be written to the predefined table

        Returns
        -------
            None
        """
        input_cols = list(log_message.columns)
        cols_write = list(set(self.columns) & set(input_cols))
        cols_null = list(set(self.columns) - set(cols_write))
        print_(f"\n{cols_write = }\n{cols_null = }", color='BOLD')

        df_write = pd.DataFrame()
        df_write[cols_write] = log_message[cols_write]
        df_write[cols_null] = ''

        if 'date' in cols_write:
            df_write['date'] = pd.to_datetime(df_write['date']).strftime("%Y-%m-%d %H:%M:%S")
        else:
            df_write['date'] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        df_write.reset_index(drop=True, inplace=True)
        self.db_connection.insert(data=df_write, table_name=self.table_name)

    def print_info(self):
        print_(f"\ntable_name: {self.table_name}"
               f"\ncolumns: {self.columns}"
               f"\ncolumns types: \n{self.column_types}", color='Magenta')



# Copyright 2023 Eurobios
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains the main class to be used
It solely contains the class ConnectDatabase
"""

import io
import logging
import typing
from typing import Union, Iterable, Dict

import pandas as pd
from sqlalchemy import exc, text

from acrocord.misc import execution
from acrocord.utils import auxiliaries
from acrocord.utils import connect
from acrocord.utils import insert, cache
from acrocord.utils.types import warning_type

LOGGER = logging.getLogger(__name__)
logging.basicConfig(format='%(levelname)s:%(message)s')


class ConnectDatabase:
    def __init__(self, cache_data=False):
        self.engine = None
        self.connection = None
        self.replace = self.rename
        self.rename_ = {}
        self.active_execution_time = False
        self.active_sql_printing = False
        self.verbose = 1
        self.cache = cache_data

    def connect(self, verbose: int = 1,
                connection: Union[dict, str] = None,
                **kwargs) -> "ConnectDatabase":
        """
        Instantiate the connexion with the database. This function has to be
        called to access other methods.

        Return a :obj:`ConnectDatabase` object.


        Parameters
        ----------
        verbose: int,
            level of verbosity.
            - 0 nothing
            - 1 metadata
            - 2 sql command
        connection: (str or dict),
            connection parameters. If a string is
            provided the file `~/.postgresql/connections.cfg` will be loaded
            as connection parameters. If a dict is provided, it must contain
            'host', 'port', 'dbname' and 'user' keys.

        Returns
        -------
        :obj:`ConnectDatabase`
        """
        if isinstance(connection, str):
            connection = connect.load_available_connections()[connection]

        self.engine = connect.connect_psql_server(
            connection=connection, **kwargs.copy())
        self.connection = self.engine.raw_connection()
        self.active_execution_time = verbose >= 1
        self.active_sql_printing = verbose >= 2
        self.verbose = verbose
        self.username = connection["user"]

        return self

    def close(self) -> None:
        """
        Close the connection

        Returns
        -------
            None
        """
        self.engine.dispose()
        self.connection.detach()
        self.connection.close()

    # ===========================================================================
    #                           READ AND WRITE
    # ==========================================================================
    @execution.execution_time
    def read_table(
            self, table_name: str,
            columns: iter = None,
            where: str = None,
            **kwargs) -> pd.DataFrame:
        """
        Read a table from server

        Parameters
        ---------
        table_name: str
            Name of the table. Naming convention "schema.name" if no schema
            is specified the "main" schema will be used
        columns: iter
            Specific columns to request.
            If None all columns are requested.
            If errors are made or columns are missed, they will be ignored
        where: str
            Request specific rows in postgresql fashion

        Returns
        -------
            Data as :obj:`pandas.DataFrame`
        """
        table_name = f"{self._get_schema(table_name)}.{self._get_name(table_name)}"
        query = self._select(table_name, columns=columns, where=where, **kwargs)

        if self.cache:
            data_frame = cache.read(table_name, where=where,
                                    columns=columns, **kwargs)
            if len(data_frame) > 0:
                return data_frame

        copy_sql = f"COPY ({query}) TO STDOUT WITH CSV HEADER"
        connection = self.engine.connect()
        with connection.connection.cursor() as cursor:
            storage = io.StringIO()
            cursor.copy_expert(copy_sql, storage)
            storage.seek(0)
        data_frame = pd.read_csv(storage, true_values=["t"], false_values=["f"])

        if self.cache:
            cache.write(data_frame, table_name,
                        columns=columns, where=where, **kwargs)
        return data_frame

    @execution.execution_time
    def write_table(self, data: typing.Union[pd.DataFrame, "GeoDataFrame"],
                    table_name: str = "tmp",
                    primary_key=None,
                    if_exists="replace_cascade",
                    foreign_keys=None,
                    column_comments=None,
                    **kwargs) -> None:
        """
        Create table based on panda's DataFrame. This function upload the data
        on server and insert all data in the table.

        Parameters
        ----------
        data: pandas.DataFrame
            Data (GeoDataFrame or DataFrame) to insert into the database.
        table_name: str
            Name of the table. Naming convention "schema.name" if no schema
            is specified the "main" schema will be used
        primary_key: list or str
            List of string specifying all keys or single string
        if_exists: str
            In case of existing table

            - "replace_cascade" : will delete all table with dependency
            - "replace" : will delete this table. Error will raise if there are dependencies
        foreign_keys: dict
            specify foreign keys in the following fashion
            {column name in the current table : foreign table name}
            Note that if the name of the foreign key is different in the
            foreign table, use the `add_key` method.
        column_comments: dict
            Columns comments of type `{column_name: column_comment}`

        Returns
        -------
            None

        >>> from acrocord import ConnectDatabase
        >>> dataframe = pd.DataFrame({'a': [155, 20, 3],
        >>>                           'b': [11, 299, 45],
        >>>                           'c': [73, 3, 39],
        >>>                           'd': [783, 488, 739],
        >>>                           "lat": [45, 45, 45],
        >>>                           "lon": [0, 1, 2]
        >>>                       })
        >>> df = ConnectDatabase().connect(
        >>>      connection="connection-name"
        >>>      ).write_table( dataframe, "test.test")
        """
        data = data.rename(columns=str.lower)
        table_name = f"{self._get_schema(table_name)}.{self._get_name(table_name)}"
        warning_type(data, verbose=self.verbose)

        if "geometry" in data.dtypes.values and self.has_postgis_extension:
            if_exists_pg = "replace" if if_exists == "replace_cascade" else if_exists
            # set postgis extension in database if not already set

            data.to_postgis(
                con=self.engine,
                name=self._get_name(table_name),
                schema=self._get_schema(table_name),
                if_exists=if_exists_pg)
        else:
            if if_exists == "replace_cascade":
                self.drop_table(table_name, option="cascade")
                self.create_table(table_name, data.columns,
                                  data.dtypes.values.astype(str),
                                  column_comments)
            with execution.silence_stdout():
                self.insert(data, table_name)

        if primary_key is not None:
            self.add_key(table_name, primary_key)
        if foreign_keys is not None:
            for key in foreign_keys.keys():
                self.add_key(table_name, key, type_="foreign",
                             table_foreign=foreign_keys[key])

    def insert(self, data: pd.DataFrame, table_name: str,
               chunksize=1000) -> None:
        """
        This function insert data in existing postgresql table

        Parameters
        ---------
        data: pandas.DataFrame ,
            Data to insert
        table_name: str
            Name of the table. Note that the column of `data` and the columns
            of the table must be the same
        chunksize: int
            Size of chunk in insert function

        Returns
        -------
            None

        """

        insert.insert(self, data, table_name, chunksize)

    def create_insert(self, data: pd.DataFrame, table_name: str,
                      *args, **kwargs):
        """
        This function creates table if it does not exist and insert in the
        table otherwise.

        Parameters
        ----------
        data: pandas.DataFrame,
            Data to insert.
        table_name: str
            Name of the table. Naming convention "schema.name" if no schema
            is specified the "main" schema will be used
        """
        if self._get_name(table_name) in self.get_table_names(
                self._get_schema(table_name)):
            self.insert(data, table_name)
        else:
            self.write_table(data, table_name, *args, **kwargs)

    def add_key(self, table_name: str, key: str, type_="primary",
                table_foreign="",
                key_foreign="") -> None:
        """
        Add key (reference between tables).

        Parameters
        ----------
        table_name: str
            The name of the table for which to add key
        key: str
            The name of the key (corresponds to a column in the table)
        type_: str
            The type of key. By default, the key type is a primary key. The
            possible values are
                - "primary" to define primary key. The primary key must be
                 composed of unique values
                - "foreign" to define foreign key. If type is set to foreign,
                the argument `table_foreign` must be provided
        table_foreign: str
            The name of the foreign table. Argument is ignored if key type
            is set to "primary".
        key_foreign: str
            The name of the foreign key (column) in the foreign table table

        Returns
        -------
            None
        """

        table_name = f"{self._get_schema(table_name)}." \
                     f"{self._get_name(table_name)}"
        try:
            if key_foreign == "":
                key_foreign = key
            if table_foreign != "":
                df_type = self.get_dtypes(table_foreign)
                typ = df_type.loc[
                    df_type["column_name"].values == key_foreign,
                    "data_type"].values[0]
                sql_cmd = f"ALTER TABLE {table_name} ALTER COLUMN " \
                          f"{key} TYPE {typ} USING {key}::{typ};"
                auxiliaries.execute_sql_cmd(self, sql_cmd, fetch=False)
            key = _iterable_or_str(key)

            sql_cmd = f"ALTER TABLE {table_name} ADD {type_.upper()} KEY {key}"
            if type_ == "foreign":
                sql_cmd += f" REFERENCES {self._get_schema(table_foreign)}." \
                           f"{self._get_name(table_foreign)} ({key_foreign});"

            auxiliaries.execute_sql_cmd(self, sql_cmd, fetch=False)
        except exc.ProgrammingError as error:
            print(error)

    # ==========================================================================
    #                          ALTER DATA
    # ==========================================================================

    def add_columns(self, table_name: str, data: pd.DataFrame,
                    index: str) -> None:
        """
        Add columns to `table_name` based on the data provided on the index
        given in argument. The name of the index column given must be the same
        in table (on postgres server) and data (the data frame object)

        Parameters
        ----------
        table_name: str
            The name of table
        data: pandas.DataFrame
            Data to add in the table.
        index:
            The name of the column in the table and the dataframe on which
            to perform the merge

        Returns
        -------
            None
        """
        table_name = f"{self._get_schema(table_name)}.{self._get_name(table_name)}"
        table_tmp = self._get_schema(table_name) + ".tmp"
        self.write_table(data, table_tmp)
        if isinstance(index, str):
            index = (index,)
        index = [id_.lower() for id_ in index]

        auxiliaries.merge(self, table_name, table_tmp,
                          out=table_name, on=index, out_type="table")
        self.drop_table(table_tmp)

    def drop_column(self,
                    table_name: str, columns: Iterable[str],
                    option: str = "cascade", type_: str = "TABLE"):
        """
        Remove some columns

        Parameters
        ----------
        table_name: str
            The name of the table to alter
        columns: iter[str]
            The columns to drop
        option: str
            Option can be either
            - 'cascade' : all dependent objects will be removed
            - '': only the columns

            Default is 'cascade'
        type_:str
            The type of object can be either
                - a table and type_="TABLE"
                - a view and type_="VIEW"
        """

        sql_cmd = f"ALTER {type_} {table_name}"
        for col in columns:
            sql_cmd += f" DROP COLUMN {col} {option},"
        auxiliaries.execute_sql_cmd(self, sql_cmd[:-1], hide=True, fetch=False)

    @auxiliaries.execute
    def drop_table(self, table_name: str, option: str = "cascade",
                   type_: str = "TABLE"):
        """
        Delete a table

        Parameters
        ----------
        table_name: str
            The name of the table to drop (delete)
        option: str
            Option can be either
                - 'cascade' : all depend objects will be removed
                - '': only the columns
            Default is 'cascade'

        type_:str
            The type of object can be either
                - a table and type_="TABLE"
                - a view and type_="VIEW"

        """
        sql_cmd = auxiliaries.drop_table(table_name, option, type_)
        return sql_cmd

    @auxiliaries.execute
    def rename(self, table_name: str, new_table_name: str,
               type_: str = "TABLE"):
        """
        Rename a table

        Parameters
        ----------
        table_name: str
            The current name of the table
        new_table_name: str
            The new name of the table
        type_:str
            The type of table, either
                - a table and type_="TABLE"
                - a view and type_="VIEW"
            Default is "TABLE"

        Returns
        -------
            SQL query
        """

        sql_cmd = f"ALTER {type_} {table_name} " \
                  f"RENAME TO {self._get_name(new_table_name)}"
        return sql_cmd

    @auxiliaries.execute
    def copy(self, table_name: str, new_table_name: str, type_="TABLE") -> str:
        """
        Copy the table and create another table (or view)

        Parameters
        ----------
        table_name: str
            The name of the table to copy
        new_table_name: str
            The name of the target table
        type_: str
            The target type object, it can be either
            - "view" does not create new data
            - "table" creates new data

        Returns
        -------
            SQL query
        """
        table_name = f"{self._get_schema(table_name)}." \
                     f"{self._get_name(table_name)}"
        new_table_name = f"{self._get_schema(new_table_name)}." \
                         f"{self._get_name(new_table_name)}"
        self.drop_table(new_table_name, type_=type_)
        sql_cmd = f"CREATE {type_} {new_table_name} AS {type_} {table_name} ;"
        return sql_cmd

    # ==========================================================================
    #                           FETCH INFORMATION
    # ==========================================================================

    def get_dtypes(
            self,
            table_name: str,
            format: str = "dataframe") -> typing.Union[dict, pd.DataFrame]:
        """
        Get the column type of the table

        Parameters
        ----------
        table_name: str
            The name of the table to get type from

        format: str
            Format of the output

        Returns
        -------
           Dict or pandas.DataFrame
        """
        sql_cmd = f"select udt_name, column_name from " \
                  f"information_schema.columns where table_name=" \
                  f"'{self._get_name(table_name)}' AND table_schema=" \
                  f"'{self._get_schema(table_name)}'"
        ret = auxiliaries.execute_sql_cmd(self, sql_cmd, hide=True, fetch=True)
        df_type = pd.DataFrame(
            ret,
            columns=["data_type", "column_name"])
        return df_type

    def get_columns(self, table_name: str) -> list:
        """
        Get the list of columns of the given table name

        Parameters
        ----------
        table_name: str
            The name of the table

        Returns
        -------
            list of column names
        """
        dtypes = self.get_dtypes(table_name)
        return list(dtypes["column_name"])

    def get_metadata(self, table_name: str) -> pd.DataFrame:
        """
        Get metadata of a given table

        Parameters
        ----------
        table_name: str
            The name of the table

        Returns
        -------
            pandas.DataFrame with metadata
        """
        sql_cmd = auxiliaries.get_metadata(table_name)
        res = auxiliaries.execute_sql_cmd(self, sql_cmd, hide=True, fetch=True)
        df_meta = pd.DataFrame(res)
        df_type = self.get_dtypes(table_name)
        return pd.merge(df_meta, df_type, on='column_name')

    def get_shape(self, table_name: str) -> typing.Tuple[int, int]:
        """
        Get shape of a table

        Parameters
        ----------
        table_name:str
            The name of the table

        Returns
        -------
            tuple of row, column shape
        """
        table_name = f"{self._get_schema(table_name)}.{self._get_name(table_name)}"
        row = auxiliaries.execute_sql_cmd(
            self,
            auxiliaries.count(table_name),
            fetch=True,
            hide=True
        )[0][0]
        col = len(self.get_columns(table_name))
        return row, col

    def get_schema_names(self) -> list:
        """
        Get the names of schemas for a given connection.
        The connection is associated with a database.

        Returns
        -------
            List of schema names
        """

        with self.engine.connect() as cursor:
            res = cursor.execute(text('SELECT * FROM pg_catalog.pg_tables'))
            data_frame = pd.DataFrame(res.fetchall())
        list_ = list(set(data_frame.iloc[:, 0]))
        list_ = [schema for schema in list_ if
                 "pg" not in schema and "schema" not in schema]
        return list_

    def get_table_names(self, schema: str = "public") -> list:
        """
        Get a list of table for a given schema

        Parameters
        ----------
        schema: str
            The name of the schema

        Returns
        -------
            List of table names
        """
        res = auxiliaries.execute_sql_cmd(
            self,
            f"SELECT * FROM pg_catalog.pg_tables where schemaname='{schema}'",
            hide=True, fetch=True)
        df_names = pd.DataFrame(res)
        if len(df_names) == 0:
            return []
        return list(df_names.iloc[:, 1])

    def get_view_names(self, schema="public"):
        """
        Get a list of views for a given schema

        Parameters
        ----------
        schema: str
            The name of the schema

        Returns
        -------
            List of view names
        """
        res = auxiliaries.execute_sql_cmd(
            self, 'SELECT * FROM pg_catalog.pg_views',
            hide=True, fetch=True)
        df_names = pd.DataFrame(res)
        df_names = df_names[df_names.iloc[:, 0] == schema]
        return list(df_names.iloc[:, 1])

    def _select(self, table_name: str, columns: Iterable[str] = None,
                where: str = None,
                limit: int = None):
        """
        Select type query

        Parameters
        ----------
        table_name: str
            The name of the table on which to apply a select
        columns: iterable of str
            The name of the selected columns
        where: str
            Selection conditions according to the indexes
        limit: int
            The maximal number of row to select

        Returns
        -------
            SQL query
        """
        table_name = f"{self._get_schema(table_name)}.{self._get_name(table_name)}"
        if columns is not None:
            columns = [col.lower() for col in columns if
                       col in self.get_columns(table_name)]
            columns = pd.Series(columns).drop_duplicates(keep='first')
            query = ",\n  ".join(columns)
            query = "SELECT " + query + " FROM " + table_name
        else:
            query = "SELECT * FROM " + table_name

        if where is not None:
            query += " WHERE " + where

        if limit is not None:
            query += " LIMIT " + str(limit)
        if self.active_sql_printing:
            execution.print_(query)
        return query

    def update(self, table_name: str, where: str = "",
               column: str = "", value: str = "NULL") -> None:
        """
        Update values in table

        Parameters
        ----------
        table_name: str
            The name of the table
        where: str
            The location according to the indexes of the values to be modified
        column: str
            The column concerned
        value
            The value

        Returns
        -------
            None
        """
        if isinstance(value, str):
            value = f"'{value}'"

        sql_cmd = f"UPDATE {table_name} SET {column}={value}"
        if where != "":
            sql_cmd += f" where {where}"

        auxiliaries.execute_sql_cmd(self, sql_cmd, fetch=False)

    @staticmethod
    def _get_schema(name):
        return name.split(".")[0].lower() if "." in name else "main"

    @staticmethod
    def _get_name(name):
        return name.split(".")[1].lower() if "." in name else name.lower()

    @auxiliaries.execute
    def create_table(self, name: str, columns: Iterable[str],
                     dtypes: Iterable[str],
                     column_comments: Dict[str, str] = None) -> str:
        """

        Parameters
        ----------
        name: str
            The name of the table
        columns: iterable of str
            The names of the columns
        dtypes: iterable of str
            The types of columns
        column_comments : dict
            The comment on columns `{column1: comment 1, ...}`

        Returns
        -------
            SQL query
        """

        return auxiliaries.create_table(name, columns, dtypes, column_comments)

    @auxiliaries.execute
    def drop(self, name: str, type_: str, column_name: str) -> str:
        """
        Drop a column or object in table

        Parameters
        ----------
        name: str
            Object to alter
        type_: str

            Type of object to alter
            - VIEW
            - TABLE (default)

        column_name: str
            Column or object to drop

        Returns
        -------
            SQL query
        """
        return auxiliaries.drop(name, type_, column_name)

    @auxiliaries.execute
    def create_schema(self, name: str) -> str:
        """
        Create a schema

        Parameters
        ----------
        name: str
            Name of the schema

        Returns
        -------
            None
        """
        return auxiliaries.create_schema(name)

    def list_table(self, schema: str) -> pd.DataFrame:
        """
        List tables provided a postregsql schema

        Parameters
        ----------
        schema: str
            Schema name

        Returns
        -------
            :obj: `pandas.DataFrame` with table list and properties

        """
        return pd.DataFrame(auxiliaries.execute_sql_cmd(
            self, auxiliaries.list_table(schema),
            fetch=True))

    def execute(self, sql: str) -> None:
        """
        Provided a query as string,
        execute the query using SQLalchemy engine with psycopg2 driver

        Parameters
        ----------
        sql: str
            Some query to execute

        Returns
        -------
            None
        """
        auxiliaries.execute_sql_cmd(self, sql, hide=False)

    # ==========================================================================
    #                       Property of connexion
    # ==========================================================================
    @property
    def has_postgis_extension(self) -> bool:
        with self.engine.connect() as con:
            out_ = con.execute(
                text("SELECT COUNT(1) FROM information_schema.routines"
                     " WHERE routine_name = 'postgis_version'"))
            has_postgis = next(out_)[0]
        return has_postgis


def _iterable_or_str(arg) -> str:
    if isinstance(arg, str):
        arg = f"({arg})"
    else:
        arg = str(tuple(arg)).replace("'", "")
    return arg



# Copyright 2023 Eurobios
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.#
from typing import Type, Iterable

import pandas as pd

from acrocord import ConnectDatabase
from acrocord.utils.types import EligibleDataType


def not_nullable(data: pd.DataFrame, columns: Iterable[str]):
    """ Function that verify if a column does not contains null values """
    for column in columns:
        assert not data[column].isnull().values.any(), f"Column" \
            f" {column} " \
            f"contains null values"


def data_type(data: pd.DataFrame, columns: Iterable[str],
              dtype: Type):
    """ Function that verify if the type of column is correct """
    for column in columns:
        assert data[column].dtype == dtype, \
            f"Column {column} is {data[column].dtype}"


def eligible_data_type(data: pd.DataFrame):
    result = data.dtypes
    for column, type_ in result.items():
        msg = f'The type of the column {column} is {type_} ' \
              f'which is not an eligible type.'
        assert type_ in EligibleDataType.get_list(), msg


def quantile(data: pd.DataFrame, columns: Iterable[str], q, threshold):
    """ Function that verifies the quantile according to a threshold  """
    for column in columns:
        assert data[column].quantile(q) < threshold


def unique(data: pd.DataFrame, columns: Iterable[str]):
    """ Function that checks if the values of a column is unique """
    for column in columns:
        assert data[
            column].nunique() == data.__len__(), f"Column {column} is not unique"


def nb_unique_index(data: pd.DataFrame, columns: Iterable[str], minimum,
                    maximum):
    """ Function that checks if the number of unique objects is included in a given interval """

    for column in columns:
        msg = f"Column {column} est compris dans l'intervalle "
        assert (data[column].nunique() < maximum) and (
            data[column].nunique() >= minimum), msg


class DataConstraints:

    def __init__(self, data: pd.DataFrame = None,
                 connection: ConnectDatabase = None):
        self.data = data.copy()
        self.constraints = {}

    def add_constraint(self, constraint: dict):
        self.constraints = {**self.constraints, **constraint}

    def test(self):
        for name in self.constraints.keys():
            function, args = self.constraints[name]
            args["data"] = self.data
            function(**args)



# Copyright 2023 Eurobios
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

import os
from typing import Dict, Tuple

import pandas as pd
import pytest

from acrocord.factory.table import TableFactory
from acrocord.utils.types import EligibleDataType


class BuildingTableFactory(TableFactory):
    # name of the table
    @classmethod
    def table_name(cls) -> str:
        return 'buildings'

    # name of the schema where to store the table
    @classmethod
    def schema_name(cls) -> str:
        return 'test'

    # column format statement and documentation.
    @classmethod
    def data_definition(cls) -> Dict[str, Tuple[EligibleDataType, str]]:
        return {
            'building_id': (EligibleDataType.INTEGER, 'Identification number'),
            'architect': (EligibleDataType.STRING, 'Architect name'),
            'height': ("float16", ''),
            'construction_date': (
                EligibleDataType.DATE_TIME,
                'Construction Date of the building'),
            # 'is_constructed': (
            #     EligibleDataType.BOOLEAN, 'Does the building already construct') -> test
        }

    # this is the primary key of your table it must exist in
    # BuildingTableFactory._doc_cols
    # and as it is a row identifier, each row of your
    # table must have a unique non-null value
    @classmethod
    def id_key(cls) -> str:
        return 'building_id'

    # This is the main table where you get data from other sources
    # (possibly another TableFactory).
    # The only restriction is to
    def _create_table(self) -> None:
        df = pd.DataFrame(data={'building_id': [11, 20, 14, 34, 61],
                                'architect': ["Durand", "Blanc", "Blanc",
                                              "Dubois", "Martin"],
                                'height': [14.4, 24.4, 35.3, 12.3, 14.4],
                                'construction_date': ['10/03/1957',
                                                      '30/11/2087',
                                                      '01/02/2070',
                                                      '04/01/1989',
                                                      '28/10/2003']
                                })
        df['construction_date'] = pd.to_datetime(df.construction_date,
                                                 format='%d/%m/%Y')
        from datetime import datetime
        df['is_constructed'] = df.construction_date < datetime.now()

        self._set_table(df)  # Required line to store the result

    @classmethod
    def get_db_connection(cls):
        return cls.connection


@pytest.fixture
def get_table_factory(get_connection) -> BuildingTableFactory:
    BuildingTableFactory.connection = get_connection
    table_factory = BuildingTableFactory(verbose=True)
    return table_factory


def test_write(get_table_factory: BuildingTableFactory, get_connection) -> None:

    get_table_factory.write_table()
    get_table_factory.get_table()
    assert "buildings" in get_connection.get_table_names("test")


def test_read(get_table_factory: BuildingTableFactory) -> None:
    get_table_factory.write_table()
    get_table_factory.read_table()


def test_implement_foreign_keys(
        get_table_factory: BuildingTableFactory) -> None:
    get_table_factory.add_foreign_keys()


def test_get_table_name(get_table_factory: BuildingTableFactory) -> None:
    assert "test.buildings" == get_table_factory.get_full_name()


def test_write_to_excel(get_table_factory: BuildingTableFactory) -> None:
    import tempfile
    get_table_factory.write_to_excel()
    assert "buildings.xlsx" in os.listdir(tempfile.gettempdir())


def test_columns(get_table_factory: BuildingTableFactory) -> None:
    print(get_table_factory.columns)
    for c in get_table_factory.columns:
        assert c in ['building_id', 'architect', 'height',
                     'construction_date',
                     'is_constructed'], f"column {c} is not in columns"
    assert isinstance(get_table_factory.columns,
                      list), "columns attribute is not a list"


def test_get_instance(get_table_factory):
    inst = get_table_factory.get_instance()
    assert inst._table is None

    inst.get_table()


def test_check_table_doc(get_table_factory):
    get_table_factory.check_table_doc(get_table_factory.get_table(),
                                      get_table_factory.data_definition())

# Test to add
# TODO:  test if deployed and _table is freed
# TODO:  test if second _get_data is faster



# Copyright 2023 Eurobios
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.#
import getpass
import glob
import os
import subprocess
import sys
from os.path import expanduser, abspath

import asyncpg
import pandas as pd
from sqlalchemy import create_engine

home = expanduser("~")
path_postgresql = f"{home}/.postgresql/"
path_acr = f"{home}/.acrocord/"


def load_available_connections():
    import configparser
    default = configparser.ConfigParser()

    connection_dict = {}
    for path in [path_acr, path_postgresql]:
        abs_path = abspath(path)

        dir_path = f"{abs_path}/connections.cfg"
        default.read(dir_path)
        connection_dict = {**connection_dict, **default._sections}

        try:
            _ = pd.DataFrame(connection_dict).drop("password").T
        except KeyError:
            pass
    for connection in connection_dict.keys():
        try:
            if connection_dict[connection]["password"] == '':
                connection_dict[connection]["password"] = ' '
        except KeyError:
            pass
    return connection_dict


def password_auth():
    path = os.path.abspath(__file__).replace(os.path.basename(__file__), "")

    cmd = "python " + path + "getpassword.py " + path_postgresql

    if not os.path.exists(path_postgresql):
        os.mkdir(path_postgresql)
    if sys.platform == "win32":
        with subprocess.Popen(cmd, shell=True,
                              stderr=subprocess.PIPE,
                              stdout=subprocess.PIPE):
            pass
    elif sys.platform == 'linux':
        subprocess.call(['xterm', "-e", cmd])
    list_of_files = glob.glob(path_postgresql + '*')
    latest_file = max(list_of_files, key=os.path.getctime)

    with open(latest_file) as file:
        password = file.read()
        file.close()
    os.remove(latest_file)
    return password


def connect_psql_server(username=getpass.getuser(), async_=False, **kwargs):

    connection_param = kwargs["connection"].copy()

    if "password" not in connection_param.keys():
        connection_param["password"] = ""

    if username != getpass.getuser():
        connection_param["user"] = username

    if connection_param[
            "password"] == "" and "sslmode" not in connection_param.keys():
        connection_param["password"] = password_auth()

    if "sslmode" in connection_param.keys():
        connect_args = {'sslmode': connection_param["sslmode"]}

    else:
        connect_args = {}

    sql_cmd = 'postgresql+psycopg2://'
    sql_cmd += connection_param["user"] + ':' + connection_param[
        "password"] + '@' + connection_param["host"] + ':'
    sql_cmd += str(connection_param["port"]) + '/' + connection_param["dbname"]
    if async_:
        return connect_async(sql_cmd)
    return create_engine(sql_cmd, connect_args=connect_args)


async def connect_async(sql_cmd):
    conn = await asyncpg.connect(
        sql_cmd.replace("postgresql+psycopg2", "postgresql"))
    return conn



# Copyright 2023 Eurobios
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.#
import datetime
import os
import sys
import time
import warnings
from contextlib import contextmanager
from functools import wraps

from memory_profiler import memory_usage


class ColorsOut:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    Red = '\033[91m'
    Green = '\033[92m'
    Blue = '\033[94m'
    Cyan = '\033[96m'
    White = '\033[97m'
    Yellow = '\033[93m'
    Magenta = '\033[95m'
    Grey = '\033[90m'
    Black = '\033[90m'
    Default = '\033[99m'
    DEFAULT = Blue
    TIME = Grey


def print_(output, color="DEFAULT"):
    print(
        f"""{ColorsOut.__getattribute__(ColorsOut, color)}{output}{ColorsOut.ENDC}""")


def execution_time(method):
    @wraps(method)
    def timed(*args, **kw):
        if sys.platform =="win32":
            return method(*args, **kw)
        starting_time = time.time()
        mem, result = memory_usage((method, args, kw), retval=True, timeout=200,
                                   interval=1e-7)
        stopping_time = time.time()

        msg = "[" + method.__name__ + "] execution time :"
        msg += "-" * (40 - len(msg)) + "  "
        msg += str(datetime.timedelta(milliseconds=(stopping_time - starting_time) * 1000))
        msg += "  " + f'Memory {int(max(mem) - min(mem))}' + " MiB"
        if len(args) > 0 and hasattr(args[0], "active_execution_time"):
            if not args[0].active_execution_time:
                return result
            else:
                print_(msg, color="TIME")
        else:
            print_(msg, color="TIME")
        return result

    return timed


@contextmanager
def silence_stdout():
    new_target = open(os.devnull, "w")
    old_target = sys.stdout
    sys.stdout = new_target
    try:
        yield new_target
    finally:
        sys.stdout = old_target


def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""

    @wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn("Call to deprecated function {}.".format(func.__name__),
                      category=DeprecationWarning,
                      stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        return func(*args, **kwargs)

    return new_func



# Copyright 2023 Eurobios
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

import pytest

from acrocord.utils.logger import Logger


@pytest.fixture
def get_logger(get_connection) -> Logger:
    logger = Logger(connection=get_connection, table_name='log_table', schema='test')
    assert 'test.log_table' == logger.table_name
    return logger


def test_write_log_dataframe(get_logger: Logger, get_example_log_dataframe) -> None:
    get_logger.write_log(get_example_log_dataframe)


def test_write_log_series(get_logger: Logger, get_example_log_series) -> None:
    get_logger.write_log(get_example_log_series)


def test_write_log_dict(get_logger: Logger, get_example_log_dict) -> None:
    get_logger.write_log(get_example_log_dict)


def test_write_log_other_type(get_logger: Logger, get_example_log_other_type) -> None:
    with pytest.raises(ValueError) as exc_info:
        get_logger.write_log(get_example_log_other_type)

        assert str(exc_info.value) == "Input data should be type of [pd.Series, pd.DataFrame, or dict]"



# Copyright 2023 Eurobios
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
from acrocord.misc import execution
import pytest


def test_deprecated():

    @execution.deprecated
    def foo():
        return "hello"

    with pytest.deprecated_call():
        foo()


def test_print():

    @execution.execution_time
    def foo():
        return "hello"

    assert foo() == "hello"



# Copyright 2023 Eurobios
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os.path
from abc import ABC
from abc import abstractmethod
from typing import Dict
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import Union

import pandas as pd
import xlsxwriter
from pandas.core.dtypes.base import ExtensionDtype
from termcolor import colored
from termcolor import cprint

from acrocord import ConnectDatabase


class TableFactory(ABC):
    """
    Description
    -----------

    Abstract class for table creation and interaction with database.

    Usage
    -----

    If you want to use the static class you can call :

    >>> table_factory = TableFactory.get_instance()

    Otherwise you can create your own instance:

    >>> table_factory= TableFactory()

    Then you can access to the table by calling

    >>> table_factory.get_table()

    If you want to save the table to the database you can call:

    >>> table_factory.write_table(...)

    Then you can read the table from the same database using:

    >>> table_factory.read_table(...)
    """
    _instance: "TableFactory" = None

    @classmethod
    @abstractmethod
    def data_definition(cls) -> Dict[str, Tuple[ExtensionDtype, str]]:
        """
        Returns
        -------
        dict
            Column datatypes and description. The dict follow this structure
            {'column_name': (dtype, ' description')}

        """

    @classmethod
    @abstractmethod
    def get_db_connection(cls):
        ...

    @classmethod
    @abstractmethod
    def table_name(cls) -> str:
        """
        Returns
        -------
        str
            the database table name
        """

    @classmethod
    @abstractmethod
    def schema_name(cls) -> str:
        """
        Returns
        -------
        str
            the database schema name
        """

    @classmethod
    def read_table(cls,
                   connection: Union[dict, str, ConnectDatabase] = None,
                   schema_name: str = None,
                   table_name: str = None,
                   columns: iter = None,
                   where: str = None
                   ) -> pd.DataFrame:
        """
        if exists, read the table from database.
        Returns
        -------
        pd.DataFrame
            the table
        """
        schema_name: str = cls.schema_name() if schema_name is None else schema_name
        table_name: str = cls.table_name() if table_name is None else table_name
        columns = list(
            cls.data_definition().keys()) if columns is None else columns

        if isinstance(connection, ConnectDatabase):
            connection_ = connection
            close_db = False
        else:
            connection_ = cls.get_db_connection()
            close_db = True
        table = connection_.read_table(
            f'{schema_name}.{table_name}', columns=columns,
            where=where)

        # assert all columns and oly columns are retrieved
        assert table.columns.isin(columns).all() and pd.Index(columns).isin(
            table.columns).all()

        # convert to correct datatype
        for col in columns:
            dtype = cls.data_definition()[col][0]
            table[col] = table[col].astype(dtype)

        if close_db:
            connection_.close()
        return table

    @classmethod
    @abstractmethod
    def id_key(cls) -> Optional[Union[str, Sequence[str]]]:
        """
        Return Primary key column name, or None
        """
        ...

    @classmethod
    def get_column_description(cls) -> dict:
        """
        Returns
        -------
        dict
            dictionary with name of column as key and its description as value
        """
        doc = cls.data_definition()
        return {col_name: col_data[1] for col_name, col_data in doc.items()}

    @classmethod
    def get_instance(cls) -> "TableFactory":
        """
        Static instance getter to avoid recomputing the table for multiple usages.
        Returns
        -------
        TableFactory
            static table factory instance
        """

        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    @classmethod
    def data_description(cls) -> pd.DataFrame:
        """

        Returns
        -------
        pd.DataFrame
            a dataframe with table documentation
        """
        return pd.DataFrame(cls.data_definition(),
                            index=['data type', 'column description']).T

    @classmethod
    def get_foreign_keys(cls) -> Dict[str, Tuple[Type['TableFactory'], str]]:
        """
        return each foreign keys to be implemented in the following format:
        {key in current TableFactory columns (str) :
        (foreign table (TableFactory), foreign key (str) )}
        """
        return {}

    @classmethod
    def get_full_name(cls):
        return f"{cls.schema_name()}.{cls.table_name()}"

    def __init__(self, verbose: bool = True):
        self._verbose: bool = verbose
        # does the table have been saved into the deployment database
        self._is_deployed: bool = False
        self._table: Optional[pd.DataFrame] = None
        self.columns = list(self.data_definition().keys())

    @abstractmethod
    def _create_table(self) -> None:
        """
        compute and set the table in the self._table attribute
        Returns
        -------
        None
        """
        self._set_table(pd.DataFrame())

    def get_table(self, columns=None) -> pd.DataFrame:
        """
        Returns
        -------
        pd.DataFrame
            table générée par la fabrique
        """
        if self._is_deployed:
            return self.read_table(self.get_db_connection(), columns=columns)
        if self._table is None:
            self._create_table()
        return self._table.copy() if columns is None else self._table[
            columns].copy()

    def _set_table(self, table) -> None:
        """
        save table in self._table variable after checking its consistency with documentation
        Parameters
        ----------
        table

        Returns
        -------

        """
        table_ = table.rename(columns=str.lower)
        data_definition = self.data_definition()
        self.check_table_doc(table_, data_definition, self.__class__)
        self._table = table_[data_definition.keys()]

    def write_table(self, db: Optional[ConnectDatabase] = None,
                    schema: str = None, table_name: str = None,
                    **kwargs) -> None:
        """
        Write table to database

        Parameters
        ----------
        db : ConnectDatabase
            the database connection
        schema : str
            name of the schema
        table_name : str
            name of the tabler
        kwargs
            additional arguments for ConnectDatabase.write_table

        Returns
        -------

        """

        if db is None:
            db = self.get_db_connection()

        schema = self.schema_name() if schema is None else schema
        table_name = self.table_name() if table_name is None else table_name
        full_name = f"{schema}.{table_name}"
        if self._verbose:
            cprint(f"[write_table] {full_name}", 'white', attrs=['reverse', ])
        data = self.get_table()
        db.write_table(data, table_name=full_name,
                       column_comments=self.get_column_description(),
                       opt_dtype=False,
                       primary_key=self.id_key(), **kwargs)

        self._is_deployed = True
        self._table = None

    @staticmethod
    def check_table_doc(table: pd.DataFrame, doc: dict, class_warning=None):
        """
        check table and documentation consistency
        Parameters
        ----------
        table
        doc
        class_warning
            the name of the class to print for warnings
        Returns
        -------

        """
        if class_warning is None:
            class_warning = TableFactory.__class__
        warning_header = f"WARNING: {class_warning} "
        doc_cols = doc.keys()
        table_cols = table.columns

        if table_cols.isin(doc_cols).all() and len(doc_cols) == len(table_cols):
            for col_name, val in doc.items():
                col_dtype, col_doc = val
                try:
                    table[col_name] = table[col_name].astype(col_dtype)
                except Exception as e:
                    cprint(f"{warning_header} {col_name}: "
                           f"cannot convert {col_name} "
                           f"from {table[col_name].dtype} to {col_dtype}.",
                           'yellow')
                    cprint(f"{e}", 'yellow')
                if col_doc is None or col_doc == "":
                    cprint(warning_header +
                           f"missing description of {col_name}.",
                           'yellow')
        else:
            doc_cols = pd.Series(doc_cols)
            table_cols = pd.Series(table_cols)

            missing_doc = table_cols[~(table_cols.isin(doc_cols))]
            if len(missing_doc) > 0:
                cprint(
                    warning_header + f" column statement is missing:\n{missing_doc}",
                    'yellow')

            missing_col = doc_cols[~(doc_cols.isin(table_cols))]
            if len(missing_col) > 0:
                cprint(warning_header + f" data is missing :\n{missing_col}",
                       'yellow')

    def write_to_excel(self, save: bool = True, path: str = None,
                       file_name: str = None, verbose=True) -> 'xlsxwriter':
        if file_name is None:
            file_name = self.table_name() + '.xlsx'
        elif len(file_name) <= 5 or file_name[-5:] != '.xlsx':
            file_name += '.xlsx'
        if path is None:
            import tempfile
            path = tempfile.gettempdir()

        data = self.get_table()
        full_path = os.path.join(path, file_name)
        writer = pd.ExcelWriter(full_path, engine='xlsxwriter')
        doc = self.data_description()[
            ['column description']].reset_index().rename(
            columns={'index': 'column name'})

        for df, sheet_name, save_index in (
                (doc, 'Column description', False), (data, 'Data', False)):
            if verbose:
                print(f"Writing '{sheet_name}' Excel sheet...", end='')
            df.to_excel(writer, sheet_name=sheet_name, index=save_index)
            if verbose:
                print(f"\rWriting '{sheet_name}' Excel sheet" + colored(" ok",
                                                                        'green'))
            if verbose:
                print("adjust column width " + sheet_name + "...", end='')
            worksheet = writer.sheets[sheet_name]
            worksheet.autofilter(0, 0, 0, len(df.columns) - 1)
            for idx, col in enumerate(df):  # loop through all columns
                series = df[col]
                max_len = max((
                    series.astype(str).map(len).max(),  # len of largest item
                    len(str(col)) + 2  # len of column name/header +2 for bold
                )) + 1  # adding a little extra space
                worksheet.set_column(idx, idx, max_len)
            if verbose:
                print("\radjust column width " + sheet_name + colored(" ok",
                                                                      'green'))
        if save:
            if verbose:
                print("save excel...", end='')
            if hasattr(writer, "save"):
                writer.save()
            else:
                writer._save()
            if verbose:
                print(
                    f"ok -> {colored(full_path, 'blue', attrs=['underline'])}")
        return writer

    def add_foreign_keys(self, db: ConnectDatabase = None):
        if db is None:
            db = self.get_db_connection()
        full_table_name = f"{self.schema_name()}.{self.table_name()}"
        for key, (
                foreign_table, foreign_key) in self.get_foreign_keys().items():
            foreign_table_full_name = f"{foreign_table.schema_name()}" \
                                      f".{foreign_table.table_name()}"
            print(
                f"{full_table_name} ({key}) -- > ({foreign_key}) "
                f"{foreign_table_full_name} ")
            db.add_key(type_='foreign',
                       table_name=full_table_name, key=key,
                       table_foreign=foreign_table_full_name,
                       key_foreign=foreign_key,
                       )
